#include <string.h>
#include <Esp.h>

#include <BLEDevice.h>
#include <BLEUtils.h>
#include <BLEServer.h>
#include <BLE2902.h>

#include "bluetooth.h"
#include "storage.h"
#include "config.h"

constexpr size_t bluetoothMaxDataSize = 200;

constexpr uint16_t DEVINFO_UUID = 0x180a;
constexpr uint16_t DEVINFO_MANUFACTURER_UUID = 0x2a29;
constexpr uint16_t DEVINFO_NAME_UUID = 0x2a24;
constexpr uint16_t DEVINFO_SERIAL_UUID = 0x2a25;

const char* deviceName = "Window";
const char* manufacturerName = "CH";
const char* serviceUUID = "22071991-feed-deaf-babe-150420870001";
const char* characteristicUUID = "22071991-feed-deaf-babe-150420870002";

bool isConnected = false;
uint8_t bluetoothOutgoingBuffer[bluetoothMaxDataSize + 1];
uint8_t* bluetoothDataBuffer;
uint8_t* bluetoothResponse;
size_t bluetoothDataCount = 0;

void bluetoothStartAdvertising();
void bluetoothDidReceiveData(uint8_t* buffer, size_t count);

class BluetoothConnection: public BLEServerCallbacks {

public:
    BluetoothConnection() {}

    void onConnect(BLEServer *server) override {
        isConnected = true;
    }

    void onDisconnect(BLEServer *server) override {
        isConnected = false;
        bluetoothStartAdvertising();
    }
};

class Receiver : public BLECharacteristicCallbacks {
    
public:
    Receiver() {}

    void onWrite(BLECharacteristic *characteristic) {
        size_t bytesReceived = characteristic->getLength();
        uint8_t* data = characteristic->getData();
        bluetoothDidReceiveData(data, bytesReceived);
    }

    void onRead(BLECharacteristic *characteristic) {
        characteristic->setValue(bluetoothOutgoingBuffer, bluetoothDataCount + 1); // Add 1 byte for response
    }
};

BluetoothConnection bluetooth{};
BLE2902 descriptor{};
BLEServer *server;
Receiver receiver{};

void bluetoothConfigure() {
    bluetoothDataBuffer = bluetoothOutgoingBuffer + 1;
    bluetoothResponse = bluetoothOutgoingBuffer;

    // Setup BLE Server
    BLEDevice::init(deviceName);
    server = BLEDevice::createServer();
    server->setCallbacks(&bluetooth);
    
    // Register message service that can receive messages and reply with a static message.
    BLEService *service = server->createService(serviceUUID);
    uint32_t properties = BLECharacteristic::PROPERTY_READ | BLECharacteristic::PROPERTY_NOTIFY | BLECharacteristic::PROPERTY_WRITE;
    BLECharacteristic* characteristicMessage = service->createCharacteristic(characteristicUUID, properties);
    characteristicMessage->setCallbacks(&receiver);
    characteristicMessage->addDescriptor(&descriptor);
    service->start();

    // Register device info service, that contains the device's UUID, manufacturer and name.
    service = server->createService(DEVINFO_UUID);
    BLECharacteristic *characteristic = service->createCharacteristic(DEVINFO_MANUFACTURER_UUID, BLECharacteristic::PROPERTY_READ);
    characteristic->setValue(manufacturerName);
    characteristic = service->createCharacteristic(DEVINFO_NAME_UUID, BLECharacteristic::PROPERTY_READ);
    characteristic->setValue(deviceName);
    characteristic = service->createCharacteristic(DEVINFO_SERIAL_UUID, BLECharacteristic::PROPERTY_READ);

    uint32_t id = ESP.getEfuseMac() >> 24;
    char chipId[9] = {0};
    sprintf(chipId, "%x", id);

    characteristic->setValue(chipId);
    service->start();
}

void bluetoothStartAdvertising() {
    // Advertise services
    BLEAdvertising *advertisement = server->getAdvertising();
    BLEAdvertisementData adv;
    adv.setName(deviceName);
    adv.setCompleteServices(BLEUUID(serviceUUID));
    advertisement->setAdvertisementData(adv);
    advertisement->start();
}

bool bluetoothIsConnected() {
    return isConnected;
}

enum class BluetoothRequest: uint8_t {

    /**
     * @brief Request the number of bytes already recorded
     * 
     * Request:
     * - No additional bytes expected
     * 
     * Response: 
     * - BluetoothResponse::success, plus the number of bytes as a uint16_t (2 bytes)
     */
    getNumberOfRecordedBytes = 0,

    /**
     * @brief Request recording data
     * 
     * Request:
     * - Bytes 1-2: Memory offset (uint16_t)
     * - Bytes 3-4: Number of bytes (uint16_t)
     * 
     * Response: 
     * - BluetoothResponse::success, plus the requested bytes
     * - BluetoothResponse::responseTooLarge if too many bytes are requested
     */
    getRecordingData = 1,

    /**
     * @brief Request deletion of recordings
     * 
     * Request:
     * - Bytes 1-2: Number of bytes to clear (uint16_t)
     * 
     * Response: 
     * - BluetoothResponse::success
     * - BluetoothResponse::invalidNumberOfBytesToDelete, if the number of bytes does not match. 
     *   This may happen when a new temperature recording is performed in between calls
     */
    clearRecordingBuffer = 2,

    /**
     * @brief Request the time since the device was turned on
     * 
     * Request:
     * - No additional bytes expected
     * 
     * Response: 
     * - BluetoothResponse::success, plus the number of seconds as a uint32_t (4 bytes)
     */
    getCurrentTime = 3,

    /**
     * @brief Request the number of seconds until the next measurement is performed
     * 
     * Request:
     * - No additional bytes expected
     * 
     * Response: 
     * - BluetoothResponse::success, plus the number of seconds as a uint16_t (2 bytes)
     */
    getSecondsUntilNextMeasurement = 4,

    /**
     * @brief Get the number of seconds 
     * 
     * Request:
     * - No additional bytes expected
     * 
     * Response: 
     * - BluetoothResponse::success, plus the number of seconds as a uint16_t (2 bytes)
     */
    getMeasurementInterval = 5,
};

enum class BluetoothResponse: uint8_t {

    /**
     * @brief The response to the last request is provided
     * 
     */
    success = 0,

    /**
     * @brief Invalid command received
     * 
     */
    invalidCommand = 1,

    responseTooLarge = 2,

    unknownCommand = 3,

    invalidNumberOfBytesToDelete = 4,
};

void setResponse(BluetoothResponse response) {
    *bluetoothResponse = static_cast<uint8_t>(response);
}

void setResponseWithoutData(BluetoothResponse response) {
    setResponse(response);
    bluetoothDataCount = 0;
}

void bluetoothSetOutgoingData(BluetoothResponse response, const uint8_t* data, size_t count) {
    if (count > bluetoothMaxDataSize) {
        setResponseWithoutData(BluetoothResponse::responseTooLarge);
        return;
    }
    memcpy(bluetoothDataBuffer, data, count);
    bluetoothDataCount = count;
    setResponse(response);
}

void setSuccessResponseWithNumber(uint16_t number) {
    bluetoothSetOutgoingData(BluetoothResponse::success, (uint8_t*) &number, sizeof(uint16_t));
}

void setSuccessResponseWithUInt32(uint32_t number) {
    bluetoothSetOutgoingData(BluetoothResponse::success, (uint8_t*) &number, sizeof(uint32_t));
}

uint16_t readNumberFromReceivedBuffer(uint8_t* buffer) {
    return *((uint16_t*) buffer + 1);
}

void bluetoothDidReceiveData(uint8_t* buffer, size_t count) {
    if (count < 1) {
        setResponseWithoutData(BluetoothResponse::invalidCommand);
        return;
    }

    uint16_t offset;
    BluetoothRequest request = static_cast<BluetoothRequest>(buffer[0]);
    switch (request)
    {
    case BluetoothRequest::getNumberOfRecordedBytes:
        count = getTotalNumberOfStoredBytes();
        setSuccessResponseWithNumber(count);
        break;

    case BluetoothRequest::clearRecordingBuffer:
        count = *((uint16_t*) buffer + 1);
        if (count != getTotalNumberOfStoredBytes()) {
            setResponseWithoutData(BluetoothResponse::invalidNumberOfBytesToDelete);
        } else {
            discardAllRecordedBytes();
            setResponseWithoutData(BluetoothResponse::success);
        }
        break;

    case BluetoothRequest::getRecordingData:
        if (count != 5) {
            setResponseWithoutData(BluetoothResponse::invalidCommand);
            break;
        }
        offset = *((uint16_t*) buffer + 1);
        count = *((uint16_t*) buffer + 3);
        if (count > bluetoothMaxDataSize) {
            setResponseWithoutData(BluetoothResponse::responseTooLarge);
            break;
        }
        count = getRecordedBytesAtOffset(bluetoothDataBuffer, offset, count);
        bluetoothDataCount = count;
        break;

    case BluetoothRequest::getCurrentTime:
        setSuccessResponseWithUInt32(time(NULL));
        break;

    case BluetoothRequest::getSecondsUntilNextMeasurement:
        count = secondsUntilNextTemperatureMeasurement();
        setSuccessResponseWithNumber(count);
        break;

    case BluetoothRequest::getMeasurementInterval:
        setSuccessResponseWithNumber(temperatureMeasurementIntervalSeconds);
        break;
    
    default:
        setResponseWithoutData(BluetoothResponse::unknownCommand);
        break;
    }
}